@@ -4,7 +4,7 @@
 /*
 The MIT License
 
-Copyright (c) 2012-2014 Denis Demidov <dennis.demidov@gmail.com>
+Copyright (c) 2012-2015 Denis Demidov <dennis.demidov@gmail.com>
 
 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
@@ -38,6 +38,7 @@ THE SOFTWARE.
 #  include <omp.h>
 #endif
 
+#include <boost/typeof/typeof.hpp>
 #include <boost/type_traits.hpp>
 #include <boost/shared_ptr.hpp>
 #include <boost/make_shared.hpp>
@@ -115,6 +116,16 @@ struct crs {
         }
     }
 
+    virtual ~crs() {}
+
+    virtual const ptr_type* ptr_data() const { return &ptr[0]; }
+    virtual const col_type* col_data() const { return &col[0]; }
+    virtual const val_type* val_data() const { return &val[0]; }
+
+    virtual ptr_type* ptr_data() { return &ptr[0]; }
+    virtual col_type* col_data() { return &col[0]; }
+    virtual val_type* val_data() { return &val[0]; }
+
     class row_iterator {
         public:
             row_iterator(
@@ -149,9 +160,9 @@ struct crs {
     };
 
     row_iterator row_begin(size_t row) const {
-        ptr_type p = ptr[row];
-        ptr_type e = ptr[row + 1];
-        return row_iterator(&col[0] + p, &col[0] + e, &val[0] + p);
+        ptr_type p = ptr_data()[row];
+        ptr_type e = ptr_data()[row + 1];
+        return row_iterator(col_data() + p, col_data() + e, val_data() + p);
     }
 
 };
@@ -173,17 +184,21 @@ crs<V, C, P> transpose(const crs<V, C, P> &A)
 
     boost::fill(T.ptr, P());
 
+    const P* Aptr = A.ptr_data();
+    const C* Acol = A.col_data();
+    const V* Aval = A.val_data();
+
     for(size_t j = 0; j < nnz; ++j)
-        ++( T.ptr[A.col[j] + 1] );
+        ++( T.ptr[Acol[j] + 1] );
 
     boost::partial_sum(T.ptr, T.ptr.begin());
 
     for(size_t i = 0; i < n; i++) {
-        for(P j = A.ptr[i], e = A.ptr[i + 1]; j < e; ++j) {
-            P head = T.ptr[A.col[j]]++;
+        for(P j = Aptr[i], e = Aptr[i + 1]; j < e; ++j) {
+            P head = T.ptr[Acol[j]]++;
 
             T.col[head] = static_cast<C>(i);
-            T.val[head] = A.val[j];
+            T.val[head] = Aval[j];
         }
     }
 
@@ -194,9 +209,15 @@ crs<V, C, P> transpose(const crs<V, C, P> &A)
 }
 
 /// Matrix-matrix product.
-template < typename V, typename C, typename P >
-crs<V, C, P> product(const crs<V, C, P> &A, const crs<V, C, P> &B) {
-    typedef typename crs<V, C, P>::row_iterator row_iterator;
+template <class MatrixA, class MatrixB>
+crs< typename value_type<MatrixA>::type >
+product(const MatrixA &A, const MatrixB &B, bool sort = false) {
+    typedef typename value_type<MatrixA>::type  V;
+    typedef ptrdiff_t C;
+    typedef ptrdiff_t P;
+
+    typedef typename row_iterator<MatrixA>::type Aiterator;
+    typedef typename row_iterator<MatrixB>::type Biterator;
 
     const size_t n = rows(A);
     const size_t m = cols(B);
@@ -224,8 +245,8 @@ crs<V, C, P> product(const crs<V, C, P> &A, const crs<V, C, P> &B) {
 #endif
 
         for(size_t ia = chunk_start; ia < chunk_end; ++ia) {
-            for(row_iterator a = A.row_begin(ia); a; ++a) {
-                for(row_iterator b = B.row_begin(a.col()); b; ++b) {
+            for(Aiterator a = A.row_begin(ia); a; ++a) {
+                for(Biterator b = B.row_begin(a.col()); b; ++b) {
                     if (marker[b.col()] != static_cast<C>(ia)) {
                         marker[b.col()]  = static_cast<C>(ia);
                         ++( c.ptr[ia + 1] );
@@ -248,11 +269,11 @@ crs<V, C, P> product(const crs<V, C, P> &A, const crs<V, C, P> &B) {
             P row_beg = c.ptr[ia];
             P row_end = row_beg;
 
-            for(row_iterator a = A.row_begin(ia); a; ++a) {
+            for(Aiterator a = A.row_begin(ia); a; ++a) {
                 C ca = a.col();
                 V va = a.value();
 
-                for(row_iterator b = B.row_begin(ca); b; ++b) {
+                for(Biterator b = B.row_begin(ca); b; ++b) {
                     C cb = b.col();
                     V vb = b.value();
 
@@ -266,6 +287,12 @@ crs<V, C, P> product(const crs<V, C, P> &A, const crs<V, C, P> &B) {
                     }
                 }
             }
+
+            if (sort) {
+                amgcl::detail::sort_row(
+                        &c.col[row_beg], &c.val[row_beg], row_end - row_beg
+                        );
+            }
         }
     }
 
@@ -297,12 +324,15 @@ std::vector<V> diagonal(const crs<V, C, P> &A, bool invert = false)
 template < typename V, typename C, typename P >
 void sort_rows(crs<V, C, P> &A) {
     const size_t n = rows(A);
+    BOOST_AUTO(Aptr, A.ptr_data());
+    BOOST_AUTO(Acol, A.col_data());
+    BOOST_AUTO(Aval, A.val_data());
 
 #pragma omp parallel for
     for(ptrdiff_t i = 0; i < static_cast<ptrdiff_t>(n); ++i) {
-        P beg = A.ptr[i];
-        P end = A.ptr[i + 1];
-        amgcl::detail::sort_row(A.col.data() + beg, A.val.data() + beg, end - beg);
+        P beg = Aptr[i];
+        P end = Aptr[i + 1];
+        amgcl::detail::sort_row(Acol + beg, Aval + beg, end - beg);
     }
 }
 
@@ -325,7 +355,7 @@ crs<V, C, P> inverse(const crs<V, C, P> &A) {
         for(row_iterator a = A.row_begin(i); a; ++a)
             Ainv.val[i * n + a.col()] = a.value();
 
-    amgcl::detail::inverse(n, Ainv.val.data());
+    amgcl::detail::inverse(n, &Ainv.val[0]);
 
     Ainv.ptr[0] = 0;
     for(size_t i = 0, idx = 0; i < n; ) {
@@ -357,6 +387,7 @@ struct builtin {
     struct params {
         params() {}
         params(const boost::property_tree::ptree&) {}
+        void get(boost::property_tree::ptree&, const std::string&) const {}
     };
 
     static std::string name() { return "builtin"; }
@@ -371,7 +402,7 @@ struct builtin {
 
     /// Copy vector to builtin backend.
     static boost::shared_ptr<vector>
-    copy_vector(const vector x, const params&)
+    copy_vector(const vector &x, const params&)
     {
         return boost::make_shared<vector>(x);
     }
@@ -394,7 +425,7 @@ struct builtin {
     struct gather {
         std::vector<ptrdiff_t> I;
 
-        gather(size_t /*src_size*/, const std::vector<ptrdiff_t> &I, const params&)
+        gather(size_t /*size*/, const std::vector<ptrdiff_t> &I, const params&)
             : I(I) { }
 
         template <class InVec, class OutVec>
@@ -404,6 +435,19 @@ struct builtin {
         }
     };
 
+    struct scatter {
+        std::vector<ptrdiff_t> I;
+
+        scatter(size_t /*size*/, const std::vector<ptrdiff_t> &I, const params&)
+            : I(I) { }
+
+        template <class InVec, class OutVec>
+        void operator()(const InVec &vals, OutVec &vec) const {
+            for(size_t i = 0; i < I.size(); ++i)
+                vec[I[i]] = vals[i];
+        }
+    };
+
     /// Create direct solver for coarse level
     static boost::shared_ptr<direct_solver>
     create_solver(boost::shared_ptr<matrix> A, const params&) {
@@ -415,7 +459,7 @@ template <class T>
 struct is_builtin_vector : boost::false_type {};
 
 template <class V>
-struct is_builtin_vector< std::vector<V> > : boost::true_type {};
+struct is_builtin_vector< std::vector<V> > : boost::is_arithmetic<V> {};
 
 template <class Iterator>
 struct is_builtin_vector< boost::iterator_range<Iterator> > : boost::true_type {};
@@ -446,7 +490,7 @@ template < typename V, typename C, typename P >
 struct ptr_data_impl< crs<V, C, P> > {
     typedef const P* type;
     static type get(const crs<V, C, P> &A) {
-        return A.ptr.data();
+        return A.ptr_data();
     }
 };
 
@@ -454,7 +498,7 @@ template < typename V, typename C, typename P >
 struct col_data_impl< crs<V, C, P> > {
     typedef const C* type;
     static type get(const crs<V, C, P> &A) {
-        return A.col.data();
+        return A.col_data();
     }
 };
 
@@ -462,14 +506,14 @@ template < typename V, typename C, typename P >
 struct val_data_impl< crs<V, C, P> > {
     typedef const V* type;
     static type get(const crs<V, C, P> &A) {
-        return A.val.data();
+        return A.val_data();
     }
 };
 
 template < typename V, typename C, typename P >
 struct nonzeros_impl< crs<V, C, P> > {
     static size_t get(const crs<V, C, P> &A) {
-        return A.ptr.empty() ? 0 : A.ptr.back();
+        return A.nrows == 0 ? 0 : A.ptr_data()[A.nrows];
     }
 };
 
@@ -492,7 +536,8 @@ struct row_begin_impl< crs<V, C, P> > {
 template < typename V, typename C, typename P >
 struct row_nonzeros_impl< crs<V, C, P> > {
     static size_t get(const crs<V, C, P> &A, size_t row) {
-        return A.ptr[row + 1] - A.ptr[row];
+        const P *Aptr = A.ptr_data();
+        return Aptr[row + 1] - Aptr[row];
     }
 };
 
