@@ -4,8 +4,8 @@
 /*
 The MIT License
 
-Copyright (c) 2012-2014 Denis Demidov <dennis.demidov@gmail.com>
-Copyright (c) 2014, Riccardo Rossi, CIMNE (International Center for Numerical Methods in Engineering)
+Copyright (c) 2012-2015 Denis Demidov <dennis.demidov@gmail.com>
+Copyright (c) 2014-2015, Riccardo Rossi, CIMNE (International Center for Numerical Methods in Engineering)
 
 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
@@ -100,33 +100,36 @@ struct constant_deflation {
  * \sa \cite Frank2001
  */
 template <
-    class                         Backend,
-    class                         Coarsening,
-    template <class> class        Relax,
+    class LocalPrecond,
     template <class, class> class IterativeSolver,
-    class                         DirectSolver = mpi::skyline_lu<typename Backend::value_type>
+    class DirectSolver = mpi::skyline_lu<typename LocalPrecond::backend_type::value_type>
     >
 class subdomain_deflation {
     public:
-        typedef amg<
-            Backend, Coarsening, Relax
-            > AMG;
-
+        typedef typename LocalPrecond::backend_type Backend;
         typedef IterativeSolver<
             Backend, detail::mpi_inner_product
             > Solver;
 
         struct params {
-            typename AMG::params          amg;
+            typename Backend::params      backend;
+            typename LocalPrecond::params precond;
             typename Solver::params       solver;
             typename DirectSolver::params direct_solver;
 
             params() {}
+
             params(const boost::property_tree::ptree &p)
-                : AMGCL_PARAMS_IMPORT_CHILD(p, amg),
+                : AMGCL_PARAMS_IMPORT_CHILD(p, precond),
                   AMGCL_PARAMS_IMPORT_CHILD(p, solver),
                   AMGCL_PARAMS_IMPORT_CHILD(p, direct_solver)
             {}
+
+            void get(boost::property_tree::ptree &p, const std::string &path) const {
+                AMGCL_PARAMS_EXPORT_CHILD(p, path, precond);
+                AMGCL_PARAMS_EXPORT_CHILD(p, path, solver);
+                AMGCL_PARAMS_EXPORT_CHILD(p, path, direct_solver);
+            }
         };
 
         typedef typename Backend::value_type value_type;
@@ -143,8 +146,8 @@ class subdomain_deflation {
         : comm(mpi_comm),
           nrows(backend::rows(Astrip)), ndv(def_vec.dim()),
           dtype( datatype<value_type>::get() ), dv_start(comm.size + 1, 0),
-          Z( ndv ), q( Backend::create_vector(nrows, prm.amg.backend) ),
-          req(2 * comm.size)
+          Z( ndv ), master_rank(0),
+          q( Backend::create_vector(nrows, prm.backend) )
         {
             MPI_Datatype mpi_ptrdiff_t = mpi::datatype<ptrdiff_t>::get();
 
@@ -161,7 +164,7 @@ class subdomain_deflation {
 
             df.resize(ndv);
             dx.resize(nz);
-            dd = Backend::create_vector(nz, prm.amg.backend);
+            dd = Backend::create_vector(nz, prm.backend);
 
             boost::shared_ptr<build_matrix> aloc = boost::make_shared<build_matrix>();
             boost::shared_ptr<build_matrix> arem = boost::make_shared<build_matrix>();
@@ -180,7 +183,7 @@ class subdomain_deflation {
                 for(int j = 0; j < ndv; ++j) {
                     for(ptrdiff_t i = 0; i < nrows; ++i)
                         z[i] = def_vec(i, j);
-                    Z[j] = Backend::copy_vector(z, prm.amg.backend);
+                    Z[j] = Backend::copy_vector(z, prm.backend);
                 }
             }
             TOC("copy deflation vectors");
@@ -207,6 +210,14 @@ class subdomain_deflation {
                 for(row_iterator1 a = backend::row_begin(Astrip, i); a; ++a) {
                     ptrdiff_t c = a.col();
 
+#ifdef AMGCL_SANITY_CHECK
+                    mpi::precondition(
+                            comm,
+                            c >= 0 && c < domain.back(),
+                            "Column number is out of bounds"
+                            );
+#endif
+
                     // Domain the column belongs to
                     ptrdiff_t d = boost::upper_bound(domain, c) - domain.begin() - 1;
 
@@ -250,7 +261,7 @@ class subdomain_deflation {
                     );
 
             MPI_Allgather(
-                    num_recv.data(),    comm.size, mpi_ptrdiff_t,
+                    &num_recv[0],    comm.size, mpi_ptrdiff_t,
                     comm_matrix.data(), comm.size, mpi_ptrdiff_t,
                     comm
                     );
@@ -272,7 +283,7 @@ class subdomain_deflation {
             recv.val.resize(rc.size());
             recv.req.resize(rnbr);
 
-            dv = Backend::create_vector( rc.size(), prm.amg.backend );
+            dv = Backend::create_vector( rc.size(), prm.backend );
 
             send.nbr.reserve(snbr);
             send.ptr.reserve(snbr + 1);
@@ -369,8 +380,8 @@ class subdomain_deflation {
             TOC("second pass");
 
             /* Finish communication pattern setup. */
-            MPI_Waitall(recv.req.size(), recv.req.data(), MPI_STATUSES_IGNORE);
-            MPI_Waitall(send.req.size(), send.req.data(), MPI_STATUSES_IGNORE);
+            MPI_Waitall(recv.req.size(), &recv.req[0], MPI_STATUSES_IGNORE);
+            MPI_Waitall(send.req.size(), &send.req[0], MPI_STATUSES_IGNORE);
 
             // Shift columns to send to local numbering:
             BOOST_FOREACH(ptrdiff_t &c, send_col) c -= chunk_start;
@@ -412,7 +423,7 @@ class subdomain_deflation {
                         &zsend[ndv * send.ptr[i]], ndv * (send.ptr[i+1] - send.ptr[i]),
                         dtype, send.nbr[i], tag_exc_vals, comm, &send.req[i]);
 
-            MPI_Waitall(recv.req.size(), recv.req.data(), MPI_STATUSES_IGNORE);
+            MPI_Waitall(recv.req.size(), &recv.req[0], MPI_STATUSES_IGNORE);
 
             boost::fill(marker, -1);
 
@@ -448,25 +459,34 @@ class subdomain_deflation {
             az->ptr.front() = 0;
             TOC("A*Z");
 
-            MPI_Waitall(send.req.size(), send.req.data(), MPI_STATUSES_IGNORE);
+            MPI_Waitall(send.req.size(), &send.req[0], MPI_STATUSES_IGNORE);
 
             /* Build deflated matrix E. */
             TIC("assemble E");
             // Who is responsible for solution of coarse problem
-            nmasters = std::min(comm.size, DirectSolver::comm_size(nz));
-            nslaves  = (comm.size + nmasters - 1) / nmasters;
-
-            master = comm.rank / nslaves;
-
-            if (comm.rank < nmasters) {
-                slaves.resize(nmasters + 1, 0);
-
-                for(int p = 0; p <= nmasters; ++p)
-                    slaves[p] = std::min(p * nslaves, comm.size);
-
-                nslaves = slaves[comm.rank + 1] - slaves[comm.rank];
+            int nmasters = std::min(comm.size, DirectSolver::comm_size(nz));
+            int nslaves  = (comm.size + nmasters - 1) / nmasters;
+            int cgroup_beg = 0, cgroup_end = 0;
+            std::vector<int> slaves(nmasters + 1, 0);
+            for(int p = 1; p <= nmasters; ++p) {
+                slaves[p] = std::min(p * nslaves, comm.size);
+                if (slaves[p-1] <= comm.rank && comm.rank < slaves[p]) {
+                    cgroup_beg  = slaves[p - 1];
+                    cgroup_end  = slaves[p];
+                    master_rank = slaves[p-1];
+                    nslaves     = cgroup_end - cgroup_beg;
+                }
             }
 
+            // Communicator for masters (used to solve the coarse problem):
+            MPI_Comm_split(comm,
+                    comm.rank == master_rank ? 0 : MPI_UNDEFINED,
+                    comm.rank, &masters_comm
+                    );
+
+            // Communicator for slaves (used to send/recv coarse data):
+            MPI_Comm_split(comm, master_rank, comm.rank, &slaves_comm);
+
             // Count nonzeros in E.
             std::vector<int> eptr(ndv + 1, 0);
             for(int j = 0; j < comm.size; ++j) {
@@ -479,20 +499,27 @@ class subdomain_deflation {
                 }
             }
 
-            std::vector<int> Eptr;
-
-            if (comm.rank < nmasters) {
-                Eptr.resize(dv_start[slaves[comm.rank+1]] - dv_start[slaves[comm.rank]] + 1);
-
-                for(int p = slaves[comm.rank], offset = dv_start[p]; p < slaves[comm.rank + 1]; ++p) {
-                    int begin = dv_start[p] - offset + 1;
-                    int size  = dv_start[p + 1] - dv_start[p];
-                    MPI_Irecv(&Eptr[begin], size, MPI_INT, p, tag_exc_lnnz, comm, &req[p]);
-                }
+            sstart.resize(nslaves);
+            ssize.resize(nslaves);
+            for(int p = cgroup_beg, i = 0, offset = dv_start[p]; p < cgroup_end; ++p, ++i) {
+                sstart[i] = dv_start[p] - offset;
+                ssize[i]  = dv_start[p + 1] - dv_start[p];
             }
 
-            MPI_Send(&eptr[1], ndv, MPI_INT, master, tag_exc_lnnz, comm);
+
+            std::vector<int> Eptr;
+            if (comm.rank == master_rank) {
+                Eptr.resize(dv_start[cgroup_end] - dv_start[cgroup_beg] + 1, 0);
+            }
+
+            MPI_Gatherv(
+                    &eptr[1], ndv, MPI_INT, &Eptr[0] + 1,
+                    const_cast<int*>(&ssize[0]), const_cast<int*>(&sstart[0]),
+                    MPI_INT, 0, slaves_comm
+                    );
+
             boost::partial_sum(eptr, eptr.begin());
+            boost::partial_sum(Eptr, Eptr.begin());
 
             // Build local strip of E.
             boost::multi_array<value_type, 2> erow(boost::extents[ndv][nz]);
@@ -530,36 +557,32 @@ class subdomain_deflation {
             // Exchange strips of E.
             std::vector<int>        Ecol;
             std::vector<value_type> Eval;
-            if (comm.rank < nmasters) {
-                MPI_Waitall(nslaves, &req[slaves[comm.rank]], MPI_STATUSES_IGNORE);
-                boost::partial_sum(Eptr, Eptr.begin());
-
+            if (comm.rank == master_rank) {
                 Ecol.resize(Eptr.back());
                 Eval.resize(Eptr.back());
 
-                for(int p = slaves[comm.rank], offset = dv_start[p]; p < slaves[comm.rank + 1]; ++p) {
-                    int begin = Eptr[dv_start[p]     - offset];
-                    int size  = Eptr[dv_start[p + 1] - offset] - begin;
-
-                    MPI_Irecv(&Ecol[begin], size, MPI_INT, p, tag_exc_dmat, comm, &req[2 * p + 0]);
-                    MPI_Irecv(&Eval[begin], size, dtype,   p, tag_exc_dmat, comm, &req[2 * p + 1]);
+                for(int p = cgroup_beg, i = 0, offset = dv_start[p]; p < cgroup_end; ++p, ++i) {
+                    sstart[i] = Eptr[dv_start[p]     - offset];
+                    ssize[i]  = Eptr[dv_start[p + 1] - offset] - sstart[i];
                 }
             }
 
-            MPI_Send(ecol.data(), ecol.size(), MPI_INT, master, tag_exc_dmat, comm);
-            MPI_Send(eval.data(), eval.size(), dtype,   master, tag_exc_dmat, comm);
+            MPI_Gatherv(
+                    &ecol[0], ecol.size(), MPI_INT, &Ecol[0],
+                    const_cast<int*>(&ssize[0]), const_cast<int*>(&sstart[0]),
+                    MPI_INT, 0, slaves_comm
+                    );
+
+            MPI_Gatherv(
+                    &eval[0], eval.size(), dtype, &Eval[0],
+                    const_cast<int*>(&ssize[0]), const_cast<int*>(&sstart[0]),
+                    dtype, 0, slaves_comm
+                    );
             TOC("assemble E");
 
             // Prepare E factorization.
             TIC("factorize E");
-            MPI_Comm_split(comm,
-                    comm.rank < nmasters ? 0 : MPI_UNDEFINED,
-                    comm.rank, &masters_comm
-                    );
-
-            if (comm.rank < nmasters) {
-                MPI_Waitall(2 * nslaves, &req[2 * slaves[comm.rank]], MPI_STATUSES_IGNORE);
-
+            if (comm.rank == master_rank) {
                 E = boost::make_shared<DirectSolver>(
                         masters_comm, Eptr.size() - 1, Eptr, Ecol, Eval, prm.direct_solver
                         );
@@ -571,27 +594,41 @@ class subdomain_deflation {
 
             TOC("setup deflation");
 
-            // Create local AMG preconditioner.
-            P = boost::make_shared<AMG>( *aloc, prm.amg );
+            // Create local preconditioner.
+            P = boost::make_shared<LocalPrecond>( *aloc, prm.precond, prm.backend );
 
             // Create iterative solver instance.
             solve = boost::make_shared<Solver>(
-                    nrows, prm.solver, prm.amg.backend,
+                    nrows, prm.solver, prm.backend,
                     detail::mpi_inner_product(mpi_comm)
                     );
 
             // Move matrices to backend.
-            Arem = Backend::copy_matrix(arem, prm.amg.backend);
-            AZ   = Backend::copy_matrix(az,   prm.amg.backend);
+            Arem = Backend::copy_matrix(arem, prm.backend);
+            AZ   = Backend::copy_matrix(az,   prm.backend);
 
             // Columns gatherer. Will retrieve columns to send from backend.
             gather = boost::make_shared<typename Backend::gather>(
-                    nrows, send_col, prm.amg.backend);
+                    nrows, send_col, prm.backend);
+
+            // Prepare Gatherv configuration for coarse solve
+            for(int p = cgroup_beg, i = 0, offset = dv_start[p]; p < cgroup_end; ++p, ++i) {
+                sstart[i] = dv_start[p] - offset;
+                ssize[i]  = dv_start[p + 1] - dv_start[p];
+            }
+
+            mstart.resize(nmasters);
+            msize.resize(nmasters);
+            for(int p = 0; p < nmasters; ++p) {
+                mstart[p] = dv_start[slaves[p]];
+                msize[p]  = dv_start[slaves[p+1]] - mstart[p];
+            }
         }
 
         ~subdomain_deflation() {
             E.reset();
             if (masters_comm != MPI_COMM_NULL) MPI_Comm_free(&masters_comm);
+            if (slaves_comm  != MPI_COMM_NULL) MPI_Comm_free(&slaves_comm);
         }
 
         template <class Vec1, class Vec2>
@@ -612,7 +649,7 @@ class subdomain_deflation {
         void residual(const Vec1 &f, const Vec2 &x, Vec3 &r) const {
             TIC("top/residual");
             start_exchange(x);
-            backend::residual(f, P->top_matrix(), x, r);
+            backend::residual(f, P->system_matrix(), x, r);
 
             finish_exchange();
 
@@ -638,17 +675,17 @@ class subdomain_deflation {
 
         boost::shared_ptr<matrix> Arem;
 
-        boost::shared_ptr<AMG>    P;
-        boost::shared_ptr<Solver> solve;
+        boost::shared_ptr<LocalPrecond> P;
+        boost::shared_ptr<Solver>       solve;
 
         mutable std::vector<value_type> df, dx, cf, cx;
         std::vector<ptrdiff_t> dv_start;
+        std::vector<int> sstart, ssize, mstart, msize;
 
         std::vector< boost::shared_ptr<vector> > Z;
 
-        MPI_Comm masters_comm;
-        int nmasters, nslaves, master;
-        std::vector<int> slaves;
+        MPI_Comm masters_comm, slaves_comm;
+        int master_rank;
         boost::shared_ptr<DirectSolver> E;
 
         boost::shared_ptr<matrix> AZ;
@@ -658,8 +695,6 @@ class subdomain_deflation {
 
         boost::shared_ptr< typename Backend::gather > gather;
 
-        mutable std::vector<MPI_Request> req;
-
         struct {
             std::vector<ptrdiff_t> nbr;
             std::vector<ptrdiff_t> ptr;
@@ -681,7 +716,7 @@ class subdomain_deflation {
             TIC("top/spmv");
 
             start_exchange(x);
-            backend::spmv(alpha, P->top_matrix(), x, beta, y);
+            backend::spmv(alpha, P->system_matrix(), x, beta, y);
 
             finish_exchange();
 
@@ -759,45 +794,38 @@ class subdomain_deflation {
         }
 
         void finish_exchange() const {
-            MPI_Waitall(recv.req.size(), recv.req.data(), MPI_STATUSES_IGNORE);
-            MPI_Waitall(send.req.size(), send.req.data(), MPI_STATUSES_IGNORE);
+            MPI_Waitall(recv.req.size(), &recv.req[0], MPI_STATUSES_IGNORE);
+            MPI_Waitall(send.req.size(), &send.req[0], MPI_STATUSES_IGNORE);
         }
 
         void coarse_solve(std::vector<value_type> &f, std::vector<value_type> &x) const
         {
             TIC("coarse solve");
-            if (comm.rank < nmasters) {
-                for(int p = slaves[comm.rank], offset = dv_start[p]; p < slaves[comm.rank + 1]; ++p) {
-                    ptrdiff_t begin = dv_start[p] - offset;
-                    ptrdiff_t size  = dv_start[p + 1] - dv_start[p];
-                    MPI_Irecv(&cf[begin], size, dtype, p, tag_exc_dvec, comm, &req[p]);
-                }
-            }
-
-            MPI_Send(f.data(), f.size(), dtype, master, tag_exc_dvec, comm);
-
-            if (comm.rank < nmasters) {
-                MPI_Waitall(nslaves, &req[slaves[comm.rank]], MPI_STATUSES_IGNORE);
+            TIC("exchange rhs");
+            MPI_Gatherv(
+                    &f[0], f.size(), dtype, &cf[0],
+                    const_cast<int*>(&ssize[0]), const_cast<int*>(&sstart[0]),
+                    dtype, 0, slaves_comm
+                    );
+            TOC("exchange rhs");
 
+            if (comm.rank == master_rank) {
                 TIC("call solver");
                 (*E)(cf, cx);
                 TOC("call solver");
 
-                if (comm.rank == 0) {
-                    for(int p = 0; p < nmasters; ++p) {
-                        int begin = dv_start[slaves[p]];
-                        int size  = dv_start[slaves[p+1]] - begin;
-                        MPI_Irecv(&x[begin], size, dtype, p, tag_exc_dvec, comm, &req[p]);
-                    }
-                }
-
-                MPI_Send(cx.data(), cx.size(), dtype, 0, tag_exc_dvec, comm);
-
-                if (comm.rank == 0)
-                    MPI_Waitall(nmasters, req.data(), MPI_STATUSES_IGNORE);
+                TIC("gather result");
+                MPI_Gatherv(
+                        &cx[0], cx.size(), dtype, &x[0],
+                        const_cast<int*>(&msize[0]), const_cast<int*>(&mstart[0]),
+                        dtype, 0, masters_comm
+                        );
+                TOC("gather result");
             }
 
-            MPI_Bcast(x.data(), x.size(), dtype, 0, comm);
+            TIC("broadcast result");
+            MPI_Bcast(&x[0], x.size(), dtype, 0, comm);
+            TOC("broadcast result");
             TOC("coarse solve");
         }
 };
@@ -807,19 +835,15 @@ class subdomain_deflation {
 namespace backend {
 
 template <
-    class                         Backend,
-    class                         Coarsening,
-    template <class> class        Relax,
+    class LocalPrecond,
     template <class, class> class IterativeSolver,
-    class                         DirectSolver,
+    class DirectSolver,
     class Vec1,
     class Vec2
     >
 struct spmv_impl<
     mpi::subdomain_deflation<
-        Backend,
-        Coarsening,
-        Relax,
+        LocalPrecond,
         IterativeSolver,
         DirectSolver
         >,
@@ -828,15 +852,13 @@ struct spmv_impl<
 {
     typedef
         mpi::subdomain_deflation<
-            Backend,
-            Coarsening,
-            Relax,
+            LocalPrecond,
             IterativeSolver,
             DirectSolver
             >
         M;
 
-    typedef typename Backend::value_type V;
+    typedef typename LocalPrecond::backend_type::value_type V;
 
     static void apply(V alpha, const M &A, const Vec1 &x, V beta, Vec2 &y)
     {
@@ -845,20 +867,16 @@ struct spmv_impl<
 };
 
 template <
-    class                         Backend,
-    class                         Coarsening,
-    template <class> class        Relax,
+    class LocalPrecond,
     template <class, class> class IterativeSolver,
-    class                         DirectSolver,
+    class DirectSolver,
     class Vec1,
     class Vec2,
     class Vec3
     >
 struct residual_impl<
     mpi::subdomain_deflation<
-        Backend,
-        Coarsening,
-        Relax,
+        LocalPrecond,
         IterativeSolver,
         DirectSolver
         >,
@@ -867,15 +885,13 @@ struct residual_impl<
 {
     typedef
         mpi::subdomain_deflation<
-            Backend,
-            Coarsening,
-            Relax,
+            LocalPrecond,
             IterativeSolver,
             DirectSolver
             >
         M;
 
-    typedef typename Backend::value_type V;
+    typedef typename LocalPrecond::backend_type::value_type V;
 
     static void apply(const Vec1 &rhs, const M &A, const Vec2 &x, Vec3 &r) {
         A.residual(rhs, x, r);
