@@ -4,7 +4,7 @@
 /*
 The MIT License
 
-Copyright (c) 2012-2014 Denis Demidov <dennis.demidov@gmail.com>
+Copyright (c) 2012-2015 Denis Demidov <dennis.demidov@gmail.com>
 
 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
@@ -42,21 +42,25 @@ THE SOFTWARE.
 
 #include <amgcl/backend/interface.hpp>
 #include <amgcl/coarsening/ruge_stuben.hpp>
-#include <amgcl/coarsening/pointwise_aggregates.hpp>
 #include <amgcl/coarsening/aggregation.hpp>
 #include <amgcl/coarsening/smoothed_aggregation.hpp>
 #include <amgcl/coarsening/smoothed_aggr_emin.hpp>
 
 #include <amgcl/relaxation/gauss_seidel.hpp>
+#include <amgcl/relaxation/multicolor_gauss_seidel.hpp>
 #include <amgcl/relaxation/ilu0.hpp>
+#include <amgcl/relaxation/parallel_ilu0.hpp>
+#include <amgcl/relaxation/ilut.hpp>
 #include <amgcl/relaxation/damped_jacobi.hpp>
 #include <amgcl/relaxation/spai0.hpp>
+#include <amgcl/relaxation/spai1.hpp>
 #include <amgcl/relaxation/chebyshev.hpp>
 
 #include <amgcl/solver/cg.hpp>
 #include <amgcl/solver/bicgstab.hpp>
 #include <amgcl/solver/bicgstabl.hpp>
 #include <amgcl/solver/gmres.hpp>
+#include <amgcl/solver/detail/default_inner_product.hpp>
 
 
 namespace amgcl {
@@ -113,9 +117,13 @@ inline std::istream& operator>>(std::istream &in, type &c)
 namespace relaxation {
 enum type {
     gauss_seidel,
+    multicolor_gauss_seidel,
     ilu0,
+    parallel_ilu0,
+    ilut,
     damped_jacobi,
     spai0,
+    spai1,
     chebyshev
 };
 
@@ -124,12 +132,20 @@ inline std::ostream& operator<<(std::ostream &os, type r)
     switch (r) {
         case gauss_seidel:
             return os << "gauss_seidel";
+        case multicolor_gauss_seidel:
+            return os << "multicolor_gauss_seidel";
         case ilu0:
             return os << "ilu0";
+        case parallel_ilu0:
+            return os << "parallel_ilu0";
+        case ilut:
+            return os << "ilut";
         case damped_jacobi:
             return os << "damped_jacobi";
         case spai0:
             return os << "spai0";
+        case spai1:
+            return os << "spai1";
         case chebyshev:
             return os << "chebyshev";
         default:
@@ -142,14 +158,22 @@ inline std::istream& operator>>(std::istream &in, type &r)
     std::string val;
     in >> val;
 
-    if (val == "gauss_seidel")
+    if (val == "multicolor_gauss_seidel")
+        r = multicolor_gauss_seidel;
+    else if (val == "gauss_seidel")
         r = gauss_seidel;
     else if (val == "ilu0")
         r = ilu0;
+    else if (val == "parallel_ilu0")
+        r = parallel_ilu0;
+    else if (val == "ilut")
+        r = ilut;
     else if (val == "damped_jacobi")
         r = damped_jacobi;
     else if (val == "spai0")
         r = spai0;
+    else if (val == "spai1")
+        r = spai1;
     else if (val == "chebyshev")
         r = chebyshev;
     else
@@ -160,6 +184,384 @@ inline std::istream& operator>>(std::istream &in, type &r)
 
 } // namespace relaxation
 
+namespace detail {
+
+//---------------------------------------------------------------------------
+template <
+    class Backend,
+    class Coarsening,
+    template <class> class Relaxation,
+    class Func
+    >
+inline
+typename boost::enable_if<
+    typename backend::relaxation_is_supported<Backend, Relaxation>::type,
+    void
+>::type
+process_amg(const Func &func) {
+    typedef amgcl::amg<Backend, Coarsening, Relaxation> AMG;
+    func.template process<AMG>();
+}
+
+template <
+    class Backend,
+    class Coarsening,
+    template <class> class Relaxation,
+    class Func
+    >
+inline
+typename boost::disable_if<
+    typename backend::relaxation_is_supported<Backend, Relaxation>::type,
+    void
+>::type
+process_amg(const Func&) {
+    throw std::logic_error("The relaxation scheme is not supported by the backend");
+}
+
+//---------------------------------------------------------------------------
+template <
+    class Backend,
+    class Coarsening,
+    class Func
+    >
+inline void process_amg(
+        runtime::relaxation::type relaxation,
+        const Func &func
+        )
+{
+    switch (relaxation) {
+        case runtime::relaxation::gauss_seidel:
+            process_amg<
+                Backend,
+                Coarsening,
+                amgcl::relaxation::gauss_seidel
+                >(func);
+            break;
+        case runtime::relaxation::multicolor_gauss_seidel:
+            process_amg<
+                Backend,
+                Coarsening,
+                amgcl::relaxation::multicolor_gauss_seidel
+                >(func);
+            break;
+        case runtime::relaxation::ilu0:
+            process_amg<
+                Backend,
+                Coarsening,
+                amgcl::relaxation::ilu0
+                >(func);
+            break;
+        case runtime::relaxation::parallel_ilu0:
+            process_amg<
+                Backend,
+                Coarsening,
+                amgcl::relaxation::parallel_ilu0
+                >(func);
+            break;
+        case runtime::relaxation::ilut:
+            process_amg<
+                Backend,
+                Coarsening,
+                amgcl::relaxation::ilut
+                >(func);
+            break;
+        case runtime::relaxation::damped_jacobi:
+            process_amg<
+                Backend,
+                Coarsening,
+                amgcl::relaxation::damped_jacobi
+                >(func);
+            break;
+        case runtime::relaxation::spai0:
+            process_amg<
+                Backend,
+                Coarsening,
+                amgcl::relaxation::spai0
+                >(func);
+            break;
+        case runtime::relaxation::spai1:
+            process_amg<
+                Backend,
+                Coarsening,
+                amgcl::relaxation::spai1
+                >(func);
+            break;
+        case runtime::relaxation::chebyshev:
+            process_amg<
+                Backend,
+                Coarsening,
+                amgcl::relaxation::chebyshev
+                >(func);
+            break;
+    }
+}
+
+//---------------------------------------------------------------------------
+template <
+    class Backend,
+    class Func
+    >
+inline void process_amg(
+        runtime::coarsening::type coarsening,
+        runtime::relaxation::type relaxation,
+        const Func &func
+        )
+{
+    switch (coarsening) {
+        case runtime::coarsening::ruge_stuben:
+            process_amg<
+                Backend,
+                amgcl::coarsening::ruge_stuben
+                >(relaxation, func);
+            break;
+        case runtime::coarsening::aggregation:
+            process_amg<
+                Backend,
+                amgcl::coarsening::aggregation
+                >(relaxation, func);
+            break;
+        case runtime::coarsening::smoothed_aggregation:
+            process_amg<
+                Backend,
+                amgcl::coarsening::smoothed_aggregation
+                >(relaxation, func);
+            break;
+        case runtime::coarsening::smoothed_aggr_emin:
+            process_amg<
+                Backend,
+                amgcl::coarsening::smoothed_aggr_emin
+                >(relaxation, func);
+            break;
+    }
+}
+
+template <class Backend, class Matrix>
+struct amg_create {
+    typedef boost::property_tree::ptree params;
+    typedef typename Backend::params backend_params;
+
+    void* &handle;
+
+    const Matrix &A;
+    const params &p;
+    const backend_params &bp;
+
+    amg_create(void* &handle, const Matrix &A, const params &p, const backend_params &bp)
+        : handle(handle), A(A), p(p), bp(bp) {}
+
+    template <class AMG>
+    void process() const {
+        handle = static_cast<void*>( new AMG(A, p, bp) );
+    }
+};
+
+struct amg_destroy {
+    void *handle;
+
+    amg_destroy(void *handle) : handle(handle) {}
+
+    template <class AMG>
+    void process() const {
+        delete static_cast<AMG*>(handle);
+    }
+};
+
+template <class Vec1, class Vec2>
+struct amg_cycle {
+    void       *handle;
+
+    Vec1 const &rhs;
+    Vec2       &x;
+
+    amg_cycle(void *handle, const Vec1 &rhs, Vec2 &x)
+        : handle(handle), rhs(rhs), x(x) {}
+
+    template <class AMG>
+    void process() const {
+        static_cast<AMG*>(handle)->cycle(rhs, x);
+    }
+};
+
+template <class Vec1, class Vec2>
+struct amg_apply {
+    void       *handle;
+
+    Vec1 const &rhs;
+    Vec2       &x;
+
+    amg_apply(void *handle, const Vec1 &rhs, Vec2 &x)
+        : handle(handle), rhs(rhs), x(x) {}
+
+    template <class AMG>
+    void process() const {
+        static_cast<AMG*>(handle)->apply(rhs, x);
+    }
+};
+
+template <class Matrix>
+struct amg_system_matrix {
+    void * handle;
+    mutable const Matrix * matrix;
+
+    amg_system_matrix(void * handle) : handle(handle), matrix(0) {}
+
+    template <class AMG>
+    void process() const {
+        matrix = &(static_cast<AMG*>(handle)->system_matrix());
+    }
+};
+
+struct amg_print {
+    void * handle;
+    std::ostream &os;
+
+    amg_print(void * handle, std::ostream &os) : handle(handle), os(os) {}
+
+    template <class AMG>
+    void process() const {
+        os << *static_cast<AMG*>(handle);
+    }
+};
+
+struct amg_get_params {
+    void * handle;
+    boost::property_tree::ptree &p;
+
+    amg_get_params(void * handle, boost::property_tree::ptree &p)
+        : handle(handle), p(p) {}
+
+    template <class AMG>
+    void process() const {
+        static_cast<AMG*>(handle)->prm.get(p, "precond.");
+    }
+};
+
+} // namespace detail
+
+/// Runtime-configurable AMG preconditioner.
+template <class Backend>
+class amg : boost::noncopyable {
+    public:
+        typedef Backend backend_type;
+
+        typedef typename Backend::value_type value_type;
+        typedef typename Backend::matrix     matrix;
+        typedef typename Backend::vector     vector;
+        typedef typename Backend::params     backend_params;
+
+        typedef boost::property_tree::ptree params;
+
+        /// Constructs the AMG hierarchy.
+        /**
+         * \param A          The system matrix.
+         * \param prm        Parameters.
+         *
+         * \note The prm argument is an instance of boost::property_tree::ptree
+         * class. The structure of the property tree should copy the structure
+         * of amgcl::AMG::params struct. E.g., one could
+         \code
+         prm.put("coarsening.aggr.eps_strong", 1e-2);
+         \endcode
+         *
+         * Additionally, the property tree may contain parameters
+         * "coarsening.type" and "relaxation.type". If those are missing,
+         * default values of runtime::coarsening::smoothed_aggregation and
+         * runtime::relaxation::spai0 will be selected.
+         *
+         * Any parameters that are not relevant to the current AMG class, are
+         * silently ignored.
+         */
+        template <class Matrix>
+        amg(
+                const Matrix &A,
+                const params &prm = params(),
+                const backend_params &backend_prm = backend_params()
+           )
+          : coarsening(prm.get("coarsening.type", runtime::coarsening::smoothed_aggregation)),
+            relaxation(prm.get("relaxation.type", runtime::relaxation::spai0)),
+            handle(0)
+        {
+            runtime::detail::process_amg<Backend>(
+                    coarsening, relaxation,
+                    runtime::detail::amg_create<Backend, Matrix>(handle, A, prm, backend_prm)
+                    );
+        }
+
+        /// Destructor.
+        ~amg() {
+            runtime::detail::process_amg<Backend>(
+                    coarsening, relaxation,
+                    runtime::detail::amg_destroy(handle)
+                    );
+        }
+
+        /// Fills the property tree with the actual parameters used.
+        void get_params(boost::property_tree::ptree &p) const {
+            runtime::detail::process_amg<Backend>(
+                    coarsening, relaxation,
+                    runtime::detail::amg_get_params(handle, p)
+                    );
+        }
+
+        /// Performs single V-cycle for the given right-hand side and solution.
+        /**
+         * \param rhs Right-hand side vector.
+         * \param x   Solution vector.
+         */
+        template <class Vec1, class Vec2>
+        void cycle(const Vec1 &rhs, Vec2 &x) const {
+            runtime::detail::process_amg<Backend>(
+                    coarsening, relaxation,
+                    runtime::detail::amg_cycle<Vec1, Vec2>(handle, rhs, x)
+                    );
+        }
+
+        /// Performs single V-cycle after clearing x.
+        /**
+         * This is intended for use as a preconditioning procedure.
+         *
+         * \param rhs Right-hand side vector.
+         * \param x   Solution vector.
+         */
+        template <class Vec1, class Vec2>
+        void apply(const Vec1 &rhs, Vec2 &x) const {
+            runtime::detail::process_amg<Backend>(
+                    coarsening, relaxation,
+                    runtime::detail::amg_apply<Vec1, Vec2>(handle, rhs, x)
+                    );
+        }
+
+        /// Returns the system matrix from the finest level.
+        const matrix& system_matrix() const {
+            runtime::detail::amg_system_matrix<matrix> top(handle);
+            runtime::detail::process_amg<Backend>(
+                    coarsening, relaxation, top
+                    );
+            return *top.matrix;
+        }
+
+        /// Returns problem size at the finest level.
+        size_t size() const {
+            return backend::rows( system_matrix() );
+        }
+
+        /// Sends information about the AMG hierarchy to output stream.
+        friend std::ostream& operator<<(std::ostream &os, const amg &a)
+        {
+            runtime::detail::process_amg<Backend>(
+                    a.coarsening, a.relaxation,
+                    runtime::detail::amg_print(a.handle, os)
+                    );
+            return os;
+        }
+
+    private:
+        const runtime::coarsening::type coarsening;
+        const runtime::relaxation::type relaxation;
+
+        void *handle;
+};
+
 /// Iterative solvers.
 namespace solver {
 enum type {
@@ -208,354 +610,9 @@ inline std::istream& operator>>(std::istream &in, type &s)
 
 namespace detail {
 
-//---------------------------------------------------------------------------
-template <
-    class Backend,
-    class Coarsening,
-    template <class> class Relaxation,
-    class Func
-    >
-inline
-typename boost::enable_if<
-    typename backend::relaxation_is_supported<Backend, Relaxation>::type,
-    void
->::type
-process_amg(const Func &func) {
-    typedef amgcl::amg<Backend, Coarsening, Relaxation> AMG;
-    func.template process<AMG>();
-}
-
-template <
-    class Backend,
-    class Coarsening,
-    template <class> class Relaxation,
-    class Func
-    >
-inline
-typename boost::disable_if<
-    typename backend::relaxation_is_supported<Backend, Relaxation>::type,
-    void
->::type
-process_amg(const Func &func) {
-    throw std::logic_error("The relaxation scheme is not supported by the backend");
-}
-
-//---------------------------------------------------------------------------
-template <
-    class Backend,
-    class Coarsening,
-    class Func
-    >
-inline void process_amg(
-        runtime::relaxation::type relaxation,
-        const Func &func
-        )
-{
-    switch (relaxation) {
-        case runtime::relaxation::gauss_seidel:
-            process_amg<
-                Backend,
-                Coarsening,
-                amgcl::relaxation::gauss_seidel
-                >(func);
-            break;
-        case runtime::relaxation::ilu0:
-            process_amg<
-                Backend,
-                Coarsening,
-                amgcl::relaxation::ilu0
-                >(func);
-            break;
-        case runtime::relaxation::damped_jacobi:
-            process_amg<
-                Backend,
-                Coarsening,
-                amgcl::relaxation::damped_jacobi
-                >(func);
-            break;
-        case runtime::relaxation::spai0:
-            process_amg<
-                Backend,
-                Coarsening,
-                amgcl::relaxation::spai0
-                >(func);
-            break;
-        case runtime::relaxation::chebyshev:
-            process_amg<
-                Backend,
-                Coarsening,
-                amgcl::relaxation::chebyshev
-                >(func);
-            break;
-    }
-}
-
-//---------------------------------------------------------------------------
-template <
-    class Backend,
-    class Func
-    >
-inline void process_amg(
-        runtime::coarsening::type coarsening,
-        runtime::relaxation::type relaxation,
-        const Func &func
-        )
-{
-    switch (coarsening) {
-        case runtime::coarsening::ruge_stuben:
-            process_amg<
-                Backend,
-                amgcl::coarsening::ruge_stuben
-                >(relaxation, func);
-            break;
-        case runtime::coarsening::aggregation:
-            process_amg<
-                Backend,
-                amgcl::coarsening::aggregation<
-                    amgcl::coarsening::pointwise_aggregates
-                    >
-                >(relaxation, func);
-            break;
-        case runtime::coarsening::smoothed_aggregation:
-            process_amg<
-                Backend,
-                amgcl::coarsening::smoothed_aggregation<
-                    amgcl::coarsening::pointwise_aggregates
-                    >
-                >(relaxation, func);
-            break;
-        case runtime::coarsening::smoothed_aggr_emin:
-            process_amg<
-                Backend,
-                amgcl::coarsening::smoothed_aggr_emin<
-                    amgcl::coarsening::pointwise_aggregates
-                    >
-                >(relaxation, func);
-            break;
-    }
-}
-
-template <class Matrix>
-struct amg_create {
-    typedef boost::property_tree::ptree params;
-
-    void* &handle;
-
-    const Matrix &A;
-    const params &p;
-
-    amg_create(void* &handle, const Matrix &A, const params &p)
-        : handle(handle), A(A), p(p) {}
-
-    template <class AMG>
-    void process() const {
-        handle = static_cast<void*>( new AMG(A, p) );
-    }
-};
-
-struct amg_destroy {
-    void *handle;
-
-    amg_destroy(void *handle) : handle(handle) {}
-
-    template <class AMG>
-    void process() const {
-        delete static_cast<AMG*>(handle);
-    }
-};
-
-template <class Vec1, class Vec2>
-struct amg_cycle {
-    void       *handle;
-
-    Vec1 const &rhs;
-    Vec2       &x;
-
-    amg_cycle(void *handle, const Vec1 &rhs, Vec2 &x)
-        : handle(handle), rhs(rhs), x(x) {}
-
-    template <class AMG>
-    void process() const {
-        static_cast<AMG*>(handle)->cycle(rhs, x);
-    }
-};
-
-template <class Vec1, class Vec2>
-struct amg_apply {
-    void       *handle;
-
-    Vec1 const &rhs;
-    Vec2       &x;
-
-    amg_apply(void *handle, const Vec1 &rhs, Vec2 &x)
-        : handle(handle), rhs(rhs), x(x) {}
-
-    template <class AMG>
-    void process() const {
-        static_cast<AMG*>(handle)->apply(rhs, x);
-    }
-};
-
-template <class Matrix>
-struct amg_top_matrix {
-    void * handle;
-    mutable const Matrix * matrix;
-
-    amg_top_matrix(void * handle) : handle(handle), matrix(0) {}
-
-    template <class AMG>
-    void process() const {
-        matrix = &(static_cast<AMG*>(handle)->top_matrix());
-    }
-};
-
-struct amg_print {
-    void * handle;
-    std::ostream &os;
-
-    amg_print(void * handle, std::ostream &os) : handle(handle), os(os) {}
-
-    template <class AMG>
-    void process() const {
-        os << *static_cast<AMG*>(handle);
-    }
-};
-
-} // namespace detail
-
-/// Runtime-configurable AMG preconditioner.
-template <class Backend>
-class amg : boost::noncopyable {
-    public:
-        typedef Backend backend_type;
-
-        typedef typename Backend::value_type value_type;
-        typedef typename Backend::matrix     matrix;
-        typedef typename Backend::vector     vector;
-
-        typedef boost::property_tree::ptree params;
-
-        /// Constructs the AMG hierarchy.
-        /**
-         * \param coarsening Coarsening kind.
-         * \param relaxation Relaxation scheme.
-         * \param A          The system matrix.
-         * \param prm        Parameters.
-         *
-         * \note The prm argument is an instance of boost::property_tree::ptree
-         * class. The structure of the property tree should copy the structure
-         * of amgcl::AMG::params struct. E.g., one could
-         \code
-         prm.put("coarsening.aggr.eps_strong", 1e-2);
-         \endcode
-         * Any parameters that are not relevant to the current AMG class, are
-         * silently ignored.
-         */
-        template <class Matrix>
-        amg(
-                runtime::coarsening::type coarsening,
-                runtime::relaxation::type relaxation,
-                const Matrix &A,
-                const params &prm = params()
-           ) : coarsening(coarsening), relaxation(relaxation), handle(0)
-        {
-            runtime::detail::process_amg<Backend>(
-                    coarsening, relaxation,
-                    runtime::detail::amg_create<Matrix>(handle, A, prm)
-                    );
-        }
-
-        /// Constructs the AMG hierarchy with default coarsening and relaxation.
-        /**
-         * \param A          The system matrix.
-         * \param prm        Parameters.
-         *
-         * \note The default values for coarsening and relaxation are
-         * smoothed_aggregation and spai0 correspondingly.
-         */
-        template <class Matrix>
-        amg(const Matrix &A, const params &prm = params())
-            : coarsening(runtime::coarsening::smoothed_aggregation),
-              relaxation(runtime::relaxation::spai0),
-              handle(0)
-        {
-            runtime::detail::process_amg<Backend>(
-                    coarsening, relaxation,
-                    runtime::detail::amg_create<Matrix>(handle, A, prm)
-                    );
-        }
-
-        /// Destructor.
-        ~amg() {
-            runtime::detail::process_amg<Backend>(
-                    coarsening, relaxation,
-                    runtime::detail::amg_destroy(handle)
-                    );
-        }
-
-        /// Performs single V-cycle for the given right-hand side and solution.
-        /**
-         * \param rhs Right-hand side vector.
-         * \param x   Solution vector.
-         */
-        template <class Vec1, class Vec2>
-        void cycle(const Vec1 &rhs, Vec2 &x) const {
-            runtime::detail::process_amg<Backend>(
-                    coarsening, relaxation,
-                    runtime::detail::amg_cycle<Vec1, Vec2>(handle, rhs, x)
-                    );
-        }
-
-        /// Performs single V-cycle after clearing x.
-        /**
-         * This is intended for use as a preconditioning procedure.
-         *
-         * \param rhs Right-hand side vector.
-         * \param x   Solution vector.
-         */
-        template <class Vec1, class Vec2>
-        void apply(const Vec1 &rhs, Vec2 &x) const {
-            runtime::detail::process_amg<Backend>(
-                    coarsening, relaxation,
-                    runtime::detail::amg_apply<Vec1, Vec2>(handle, rhs, x)
-                    );
-        }
-
-        /// Returns the system matrix from the finest level.
-        const matrix& top_matrix() const {
-            runtime::detail::amg_top_matrix<matrix> top(handle);
-            runtime::detail::process_amg<Backend>(
-                    coarsening, relaxation, top
-                    );
-            return *top.matrix;
-        }
-
-        /// Returns problem size at the finest level.
-        size_t size() const {
-            return backend::rows( top_matrix() );
-        }
-
-        /// Sends information about the AMG hierarchy to output stream.
-        friend std::ostream& operator<<(std::ostream &os, const amg &a)
-        {
-            runtime::detail::process_amg<Backend>(
-                    a.coarsening, a.relaxation,
-                    runtime::detail::amg_print(a.handle, os)
-                    );
-            return os;
-        }
-
-    private:
-        const runtime::coarsening::type coarsening;
-        const runtime::relaxation::type relaxation;
-
-        void *handle;
-};
-
-namespace detail {
-
 template <
     class Backend,
+    class InnerProduct,
     class Func
     >
 inline void process_solver(
@@ -566,50 +623,47 @@ inline void process_solver(
     switch (solver) {
         case runtime::solver::cg:
             {
-                typedef amgcl::solver::cg<Backend> Solver;
+                typedef amgcl::solver::cg<Backend, InnerProduct> Solver;
                 func.template process<Solver>();
             }
             break;
         case runtime::solver::bicgstab:
             {
-                typedef amgcl::solver::bicgstab<Backend> Solver;
+                typedef amgcl::solver::bicgstab<Backend, InnerProduct> Solver;
                 func.template process<Solver>();
             }
             break;
         case runtime::solver::bicgstabl:
             {
-                typedef amgcl::solver::bicgstabl<Backend> Solver;
+                typedef amgcl::solver::bicgstabl<Backend, InnerProduct> Solver;
                 func.template process<Solver>();
             }
             break;
         case runtime::solver::gmres:
             {
-                typedef amgcl::solver::gmres<Backend> Solver;
+                typedef amgcl::solver::gmres<Backend, InnerProduct> Solver;
                 func.template process<Solver>();
             }
             break;
     }
 }
 
+template <class Backend>
 struct solver_create {
     typedef boost::property_tree::ptree params;
+    typedef typename Backend::params backend_params;
 
     void * &handle;
-    const params &prm;
     size_t n;
+    const params &sprm;
+    const backend_params &bprm;
 
-    solver_create(void * &handle, const params &prm, size_t n)
-        : handle(handle), prm(prm), n(n) {}
+    solver_create(void * &handle, size_t n, const params &sprm, const backend_params &bprm)
+        : handle(handle), n(n), sprm(sprm), bprm(bprm) {}
 
     template <class Solver>
     void process() const {
-        handle = static_cast<void*>(
-                new Solver(
-                    n,
-                    prm.get_child("solver", amgcl::detail::empty_ptree()),
-                    prm.get_child("amg.backend", amgcl::detail::empty_ptree())
-                    )
-                );
+        handle = static_cast<void*>(new Solver(n, sprm, bprm));
     }
 };
 
@@ -624,31 +678,42 @@ struct solver_destroy {
     }
 };
 
+struct solver_get_params {
+    void * handle;
+    boost::property_tree::ptree &p;
+
+    solver_get_params(void * handle, boost::property_tree::ptree &p)
+        : handle(handle), p(p) {}
+
+    template <class Solver>
+    void process() const {
+        static_cast<Solver*>(handle)->prm.get(p, "");
+    }
+};
+
 template <
-    class Backend,
     class Matrix,
+    class Precond,
     class Vec1,
     class Vec2
     >
 struct solver_solve {
-    typedef typename Backend::value_type value_type;
+    typedef typename Precond::backend_type::value_type value_type;
 
     void * handle;
 
-    runtime::amg<Backend> const &P;
-
-    Matrix const &A;
-    Vec1   const &rhs;
-    Vec2         &x;
+    Matrix  const &A;
+    Precond const &P;
+    Vec1    const &rhs;
+    Vec2          &x;
 
     size_t     &iters;
     value_type &resid;
 
-    solver_solve(void * handle, const runtime::amg<Backend> &P,
-            const Matrix &A, const Vec1 &rhs, Vec2 &x, size_t &iters,
-            value_type &resid
+    solver_solve(void * handle, const Matrix &A, const Precond &P,
+            const Vec1 &rhs, Vec2 &x, size_t &iters, value_type &resid
             )
-        : handle(handle), P(P), A(A), rhs(rhs), x(x),
+        : handle(handle), A(A), P(P), rhs(rhs), x(x),
           iters(iters), resid(resid)
     {}
 
@@ -661,56 +726,51 @@ struct solver_solve {
 
 } // namespace detail
 
-/// Runtime-configurable class that creates a pair of AMG preconditioner and iterative solver
-template <class Backend>
-class make_solver : boost::noncopyable {
+template <
+    class Backend,
+    class InnerProduct = amgcl::solver::detail::default_inner_product
+    >
+class iterative_solver {
     public:
-        typedef typename Backend::value_type value_type;
-        typedef boost::property_tree::ptree params;
+        typedef Backend backend_type;
+        typedef boost::property_tree::ptree   params;
+        typedef typename Backend::params backend_params;
 
-        /// Constructs the AMG hierarchy and creates iterative solver.
+        typedef typename Backend::value_type value_type;
+        typedef typename Backend::vector     vector;
+
+        /// Constructs the iterative solver.
         /**
-         * \param coarsening Coarsening kind.
-         * \param relaxation Relaxation scheme.
-         * \param solver     Iterative solver.
-         * \param A          The system matrix.
-         * \param prm        Parameters.
+         * \param n           System size.
+         * \param solver_prm  Solver parameters.
+         * \param backend_prm Backend parameters.
          *
-         * \note The prm argument is an instance of boost::property_tree::ptree
-         * class. The structure of the property tree is a union of AMG::params
-         * and Solver::params struct. E.g., one could
-         \code
-         prm.put("coarsening.aggr.eps_strong", 1e-2);
-         prm.put("tol", 1e-6);
-         \endcode
-         * Any parameters that are not relevant to the current AMG or Solver
-         * classes, are silently ignored.
+         * The solver_prm property tree may contain "type" entry that would be
+         * used for determination of iterative solver type. When left
+         * unspecified, runtime::solver::bicgstab would be used by default.
+         *
+         * Any parameters that are not relevant to the current solver classes
+         * are silently ignored.
          */
-        template <class Matrix>
-        make_solver(
-                runtime::coarsening::type coarsening,
-                runtime::relaxation::type relaxation,
-                runtime::solver::type     solver,
-                const Matrix &A,
-                const params &prm = params()
+        iterative_solver(
+                size_t n,
+                const params &solver_prm = params(),
+                const backend_params &backend_prm = backend_params()
                 )
-            : P(coarsening, relaxation, A,
-                    prm.get_child("amg", amgcl::detail::empty_ptree())
-               ),
-              solver(solver),
+            : solver(solver_prm.get("type", runtime::solver::bicgstab)),
               handle(0)
         {
-            runtime::detail::process_solver<Backend>(
+            runtime::detail::process_solver<Backend, InnerProduct>(
                     solver,
-                    runtime::detail::solver_create(
-                        handle, prm, amgcl::backend::rows(A)
+                    runtime::detail::solver_create<Backend>(
+                        handle, n, solver_prm, backend_prm
                         )
                     );
         }
 
         /// Destructor.
-        ~make_solver() {
-            runtime::detail::process_solver<Backend>(
+        ~iterative_solver() {
+            runtime::detail::process_solver<Backend, InnerProduct>(
                     solver,
                     runtime::detail::solver_destroy(handle)
                     );
@@ -719,6 +779,7 @@ class make_solver : boost::noncopyable {
         /// Solves the linear system for the given system matrix.
         /**
          * \param A   System matrix.
+         * \param P   Preconditioner.
          * \param rhs Right-hand side.
          * \param x   Solution vector.
          *
@@ -729,50 +790,51 @@ class make_solver : boost::noncopyable {
          * reasonably good preconditioner for several subsequent time steps
          * \cite Demidov2012.
          */
-        template <class Matrix, class Vec1, class Vec2>
+        template <class Matrix, class Precond, class Vec1, class Vec2>
         boost::tuple<size_t, value_type> operator()(
                 Matrix  const &A,
+                Precond const &P,
                 Vec1    const &rhs,
+#ifdef BOOST_NO_CXX11_RVALUE_REFERENCES
                 Vec2          &x
+#else
+                Vec2          &&x
+#endif
                 ) const
         {
             size_t     iters = 0;
             value_type resid = 0;
 
-            runtime::detail::process_solver<Backend>(
+            runtime::detail::process_solver<Backend, InnerProduct>(
                     solver,
-                    runtime::detail::solver_solve<Backend, Matrix, Vec1, Vec2>(
-                        handle, P, A, rhs, x, iters, resid)
+                    runtime::detail::solver_solve<Matrix, Precond, Vec1, Vec2>(
+                        handle, A, P, rhs, x, iters, resid)
                     );
 
             return boost::make_tuple(iters, resid);
         }
 
-        /// Solves the linear system for the given right-hand side.
+        /// Solves the linear system for the same matrix that was used for the preconditioner construction.
         /**
+         * \param P   Preconditioner.
          * \param rhs Right-hand side.
          * \param x   Solution vector.
          */
-        template <class Vec1, class Vec2>
+        template <class Precond, class Vec1, class Vec2>
         boost::tuple<size_t, value_type> operator()(
+                Precond const &P,
                 Vec1    const &rhs,
+#ifdef BOOST_NO_CXX11_RVALUE_REFERENCES
                 Vec2          &x
+#else
+                Vec2          &&x
+#endif
                 ) const
         {
-            return (*this)(P.top_matrix(), rhs, x);
+            return (*this)(P.system_matrix(), P, rhs, x);
         }
 
-        /// Reference to the constructed AMG hierarchy.
-        const runtime::amg<Backend>& amg() const {
-            return P;
-        }
-
-        /// Returns problem size at the finest level.
-        size_t size() const {
-            return P.size();
-        }
     private:
-        runtime::amg<Backend>        P;
         const runtime::solver::type  solver;
         void                        *handle;
 };
