@@ -4,7 +4,7 @@
 /*
 The MIT License
 
-Copyright (c) 2012-2014 Denis Demidov <dennis.demidov@gmail.com>
+Copyright (c) 2012-2015 Denis Demidov <dennis.demidov@gmail.com>
 
 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
@@ -57,6 +57,8 @@ template <
     >
 class gmres {
     public:
+        typedef Backend backend_type;
+
         typedef typename Backend::vector     vector;
         typedef typename Backend::value_type value_type;
         typedef typename Backend::params     backend_params;
@@ -83,6 +85,12 @@ class gmres {
                   AMGCL_PARAMS_IMPORT_VALUE(p, maxiter),
                   AMGCL_PARAMS_IMPORT_VALUE(p, tol)
             {}
+
+            void get(boost::property_tree::ptree &p, const std::string &path) const {
+                AMGCL_PARAMS_EXPORT_VALUE(p, path, M);
+                AMGCL_PARAMS_EXPORT_VALUE(p, path, maxiter);
+                AMGCL_PARAMS_EXPORT_VALUE(p, path, tol);
+            }
         };
 
         /// \copydoc amgcl::solver::cg::cg
@@ -128,11 +136,17 @@ class gmres {
         {
             size_t iter = 0;
 
-            value_type norm_r0 = restart(A, rhs, P, x);
-            if (norm_r0 < amgcl::detail::eps<value_type>(n))
-                return boost::make_tuple(0, 0);
+            value_type norm_rhs = norm(rhs);
+            if (norm_rhs < amgcl::detail::eps<value_type>(n)) {
+                backend::clear(x);
+                return boost::make_tuple(0, norm_rhs);
+            }
 
-            value_type eps = prm.tol * norm_r0, res_norm;
+            value_type eps = prm.tol * norm_rhs;
+
+            value_type res_norm = restart(A, rhs, P, x);
+            if (res_norm < eps)
+                return boost::make_tuple(0, res_norm / norm_rhs);
 
             do {
                 for(int i = 0; i < prm.M && iter < prm.maxiter; ++i, ++iter) {
@@ -140,7 +154,7 @@ class gmres {
 
                     if (res_norm < eps) {
                         update(x, i);
-                        return boost::make_tuple(iter + 1, res_norm / norm_r0);
+                        return boost::make_tuple(iter + 1, res_norm / norm_rhs);
                     };
                 }
 
@@ -148,7 +162,7 @@ class gmres {
                 res_norm = restart(A, rhs, P, x);
             } while (iter < prm.maxiter && res_norm > eps);
 
-            return boost::make_tuple(iter, res_norm / norm_r0);
+            return boost::make_tuple(iter, res_norm / norm_rhs);
         }
 
         /// Solves the linear system for the same matrix that was used for the AMG preconditioner construction.
@@ -164,10 +178,13 @@ class gmres {
                 Vec2          &x
                 ) const
         {
-            return (*this)(P.top_matrix(), P, rhs, x);
+            return (*this)(P.system_matrix(), P, rhs, x);
         }
-    private:
+
+    public:
         params prm;
+
+    private:
         size_t n;
 
         mutable boost::multi_array<value_type, 2> H;
